/*
    files.c -- libnfsidmap plugin using files

    Copyright (C) 2019 Anders Blomdell <anders.blomdell@gmail.com>

    libnfsidmap_files is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <errno.h>
#include <err.h>
#include <stdio.h>
#include <regex.h>

#include <nfsidmap.h>
#include <nfsidmap_plugin.h>

struct domain {
  struct domain *next;
  char *user_path;
  char *group_path;
  char *name;
  char *pattern;
  regex_t preg;
  int nmatch;
} *files_domain;

/* Traversal methods */

static int domain_struct_match_and_call(const char *name,
                                        const char *domain,
                                        int (*callback)(struct domain *domain,
                                                        char *name,
                                                        void *context),
                                        void *context)
{
  for (struct domain *d=files_domain ; d ; d=d->next) {
    IDMAP_LOG(4, ("domain_struct_match_and_call:"));
    IDMAP_LOG(4, ("  match '%s'", name));
    IDMAP_LOG(4, ("  to    '%s'", d->pattern));
    regmatch_t *pmatch = malloc(sizeof(pmatch) * d->nmatch);
    if (regexec(&d->preg, name, d->nmatch, pmatch, 0) == 0) {
      for (int i = 1 ; i < d->nmatch ; i++) {
        if (pmatch[i].rm_so != -1) {
          int namelength = pmatch[i].rm_eo - pmatch[i].rm_so;
          char *user = malloc(namelength + 1);
          memcpy(user, name + pmatch[i].rm_so, namelength);
          user[namelength] = 0;
          
          IDMAP_LOG(4, ("domain_struct_match_and_call: callback '%s'", user));
          int err = callback(d, user, context);
          free(user);
          if (err == 0) {
            free(pmatch);
            return err;
          }
        }
      }
    }
    free(pmatch);
  }
  return ENOENT;
}

static int domain_struct_call(int (*callback)(struct domain *domain,
                                              void *context),
                              void *context)
{
  for (struct domain *d=files_domain ; d ; d=d->next) {
    int err = callback(d, context);
    if (err == 0) {
      return err;
    }
  }
  return ENOENT;
}

/* Helper methods and their callbacks */

struct pwnam_args {
  struct passwd *pw;
  const char *domain_name;
  struct domain **domainp;
  char *buf;
  size_t buflen;
};

static int pwnam_callback(struct domain *domain,
                          char *name,
                          void *context)
{
  int result = ENOENT;
  struct pwnam_args *args = context;

  if (args->domain_name == NULL ||
      strcmp(args->domain_name, domain->name) == 0) {
    FILE *f = fopen(domain->user_path, "r");
    struct passwd *pwp;
    while (fgetpwent_r(f, args->pw, args->buf, args->buflen, &pwp) == 0) {
      if (strcmp(name, args->pw->pw_name) == 0) {
        if (args->domainp) {
          *args->domainp = domain;
        }
        result = 0;
        break;
      }
    }
    fclose(f);
  }
  return result;
}

static int files_getpwnam(struct passwd *pw,
                          const char *name,
                          const char *domain,
                          struct domain **domainp,
                          void **free_when_ready)
{
  int err;
  size_t buflen = sysconf(_SC_GETPW_R_SIZE_MAX);
  struct pwnam_args args = {
    .pw=pw,
    .domain_name=domain,
    .domainp=domainp,
    .buf=malloc(buflen),
    .buflen=buflen,
  };

  err = domain_struct_match_and_call(name, domain, pwnam_callback, &args);
  *free_when_ready = args.buf;
  return err;
}

struct pwuid_args {
  struct passwd *pw;
  uid_t uid;
  const char *domain_name;
  struct domain **domainp;
  char *buf;
  size_t buflen;
};

static int pwuid_callback(struct domain *domain,
                          void *context)
{
  int result = ENOENT;
  struct pwuid_args *args = context;
  struct passwd *pwp;
  
  if (args->domain_name == NULL ||
      strcmp(args->domain_name, domain->name) == 0) {
    FILE *f = fopen(domain->user_path, "r");
    while (fgetpwent_r(f, args->pw, args->buf, args->buflen, &pwp) == 0) {
      if (args->pw->pw_uid == args->uid) {
        if (args->domainp) {
          *args->domainp = domain;
        }
        result = 0;
        break;
      }
    }
    fclose(f);
  }
  return result;
}

static int files_getpwuid(struct passwd *pw,
                          uid_t uid,
                          char *domain,
                          struct domain **domainp,
                          void **free_when_ready)
{
  int err;
  size_t buflen = sysconf(_SC_GETPW_R_SIZE_MAX);
  struct pwuid_args args = {
    .pw = pw,
    .uid = uid,
    .domain_name=domain,
    .domainp=domainp,
    .buf = malloc(buflen),
    .buflen = buflen
  };

  err = domain_struct_call(pwuid_callback, &args);
  *free_when_ready = args.buf;
  return err;
}

struct grnam_args {
  struct group *gr;
  const char *domain_name;
  struct domain **domainp;
  char *buf;
  size_t buflen;
};

static int grnam_callback(struct domain *domain,
                          char *name,
                          void *context)
{
  int result = ENOENT;
  struct grnam_args *args = context;
  
  if (args->domain_name == NULL ||
      strcmp(args->domain_name, domain->name) == 0) {
    FILE *f = fopen(domain->group_path, "r");
    struct group *grp;
    while (fgetgrent_r(f, args->gr, args->buf, args->buflen, &grp) == 0) {
      if (strcmp(name, args->gr->gr_name) == 0) {
        if (args->domainp) {
          *args->domainp = domain;
        }
        result = 0;
        break;
      }
    }
    fclose(f);
  }
  return result;
}

static int files_getgrnam(struct group *gr,
                          const char *name,
                          const char *domain,
                          struct domain **domainp,
                          void **free_when_ready)
{
  size_t buflen = sysconf(_SC_GETGR_R_SIZE_MAX);
  struct grnam_args args = {
    .gr = gr,
    .domain_name=domain,
    .domainp=domainp,
    .buf = malloc(buflen),
    .buflen = buflen
  };

  int err = domain_struct_match_and_call(name, domain, grnam_callback, &args);
  *free_when_ready = args.buf;
  return err;
}

struct grgid_args {
  struct group *gr;
  uid_t gid;
  const char *domain_name;
  struct domain **domainp;
  char *buf;
  size_t buflen;
};

static int grgid_callback(struct domain *domain,
                          void *context)
{
  int result = ENOENT;
  struct grgid_args *args = context;
  struct group *grp;
  
  if (args->domain_name == NULL ||
      strcmp(args->domain_name, domain->name) == 0) {
    FILE *f = fopen(domain->group_path, "r");
    while (fgetgrent_r(f, args->gr, args->buf, args->buflen, &grp) == 0) {
      if (args->gr->gr_gid == args->gid) {
        if (args->domainp) {
          *args->domainp = domain;
        }
        result = 0;
        break;
      }
    }
    fclose(f);
  }
  return result;
}

static int files_getgrgid(struct group *gr,
                          uid_t gid,
                          char *domain,
                          struct domain **domainp,
                          void **free_when_ready)
{
  size_t buflen = sysconf(_SC_GETGR_R_SIZE_MAX);
  struct grgid_args args = {
    .gr = gr,
    .gid = gid,
    .domain_name=domain,
    .domainp=domainp,
    .buf = malloc(buflen),
    .buflen = buflen
  };

  int err = domain_struct_call(grgid_callback, &args);
  *free_when_ready = args.buf;
  return err;
}

/* Utility methods */

int files_getgrouplist(struct domain *domain,
                       const char *user,
                       gid_t group,
                       gid_t *groups,
                       int *ngroups)
{
  int maxgroups = *ngroups;
  int i = 0;

  if (i < maxgroups) {
    groups[i] = group;
  }
  i++;
  size_t buflen = sysconf(_SC_GETGR_R_SIZE_MAX);
  char *buf = malloc(buflen);

  FILE *f = fopen(domain->group_path, "r");
  struct group gr, *grp;
  while (fgetgrent_r(f, &gr, buf, buflen, &grp) == 0) {
    if (gr.gr_gid == group) {
      /* Primary group for user, skip */
      continue;
    }
    for (int j = 0; gr.gr_mem[j]; j++) {
      if (strcmp(gr.gr_mem[j], user) == 0) {
        if (i < maxgroups) {
          groups[i] = gr.gr_gid;
        }
        i++;
      }
    }
  }
  fclose(f);
  free(buf);
  *ngroups = i;
  if (i > maxgroups) {
    return -1;
  }
  return i;
}

static int write_name(char *dest, size_t len, char *localname, char *domain)
{
  if (strlen(localname) + 1 + strlen(domain) + 1 > len) {
    return ENOMEM;
  } else {
    strcpy(dest, localname);
    strcat(dest, "@");
    strcat(dest, domain);
    return 0;
  }
}

/* Interface methods */

static int files_name_to_uid(char *name, uid_t *uid)
{
  int err;
  struct passwd pw;
  void *when_ready = NULL;
  
  err = files_getpwnam(&pw, name, NULL, NULL, &when_ready);
  if (err == 0) {
    *uid = pw.pw_uid;
  }  
  free(when_ready);
  return -err;
}

static int files_name_to_gid(char *name, gid_t *gid)
{
  int err;
  void *when_ready = NULL;
  struct group gr;
  
  err = files_getgrnam(&gr, name, NULL, NULL, &when_ready);
  if (err == 0) {
    *gid = gr.gr_gid;
  }  
  free(when_ready);
  return -err;
}

static int files_uid_to_name(uid_t uid, char *domain,
                             char *name, size_t len)
{
  int err;
  void *when_ready = NULL;
  struct passwd pw;
  struct domain *domainp;
  
  err = files_getpwuid(&pw, uid, domain, &domainp, &when_ready);
  if (err == 0) {
    err = write_name(name, len, pw.pw_name, domainp->name);
  }
  free(when_ready);
  return -err;
}

static int files_gid_to_name(gid_t gid, char *domain, char *name, size_t len)
{
  int err;
  void *when_ready = NULL;
  struct group gr;
  struct domain *domainp;
  
  err = files_getgrgid(&gr, gid, domain, &domainp, &when_ready);
  if (err == 0) {
    err = write_name(name, len, gr.gr_name, domainp->name);
  }
  free(when_ready);
  return -err;
}

static char *strdup_or_null(const char *s)
{
  if (s) {
    return strdup(s);
  } else {
    return NULL;
  }
}

static const char *config_get(const char *section,
                              const char *prefix,
                              const char *suffix)
{
  const char *result = NULL;
  char *tag = malloc(strlen(prefix) + strlen(suffix) + 1);
  if (tag) {
    tag[0] = 0;
    strcat(tag, prefix);
    strcat(tag, suffix);
    result = nfsidmap_config_get("Files", tag);
    free(tag);
  }
  return result;
}

static struct domain *domain_init(char *domain)
{
  IDMAP_LOG(4, ("domain_init: %s", domain));
  struct domain *result = malloc(sizeof(*result));
  
  if (!result) {
    warn("files_init: Failed to allocate domain struct for '%s'", domain);
    goto no_result;
  }
  memset(result, 0, sizeof(*result));
  
  char *patterns = strdup_or_null(config_get("Files", "Patterns-", domain));
  if (!patterns) {
    warnx("files_init: Patterns-%s  missing", domain);
    goto no_patterns;
  }
#define SPECIAL ",%^.[$()|*+?{\\"
  /* Sanity check patterns, count number of groups and needed escapes */
  int escapes;
  int groups;
  int percents;
  char *p;
  for (percents=0, groups=1, escapes=0, p=strpbrk(patterns, SPECIAL) ;
       p ;
       p = strpbrk(p+1, SPECIAL)) {
    if (p[0] == ',') {
      if (percents != 1) {
        warnx("files_init: Group %d has wrong number of wildcards '%s' [%s]",
              groups, result->pattern, domain);
        goto invalid_pattern;
      }
      percents = 0;
      groups++;
    } else if (p[0] == '%') {
      percents++;
    } else {
      escapes++;
    }
  }
  if (percents != 1) {
    warnx("files_init: Group %d has wrong number of wildcards '%s'",
          groups, patterns);
    goto invalid_pattern;
  }
  IDMAP_LOG(4, ("domain_init: groups=%d escapes=%d patterns=%s\n",
                groups, escapes, patterns));

  /* Construct the combined regexp */
  result->pattern = malloc(strlen(patterns) +
                           strlen("^([_a-zA-Z0-9]*)$") * groups +
                           escapes + 1);
  result->pattern[0] = 0;
  strcat(result->pattern, "^");
  char *p0, *p1;
  for (p0=patterns, p1=strpbrk(patterns, SPECIAL) ;
       p1 ;
       p0=p1+1, p1 = strpbrk(p1+1, SPECIAL)) {
    strncat(result->pattern, p0, p1 - p0);
    if (p1[0] == ',') {
      strcat(result->pattern, "|");
    } else if (p1[0] == '%') {
      strcat(result->pattern, "([_a-zA-Z0-9]*)");
    } else {
      strcat(result->pattern, "\\");
      strncat(result->pattern, p1, 1);
    }
  }        
  strcat(result->pattern, p0);
  strcat(result->pattern, "$");
  IDMAP_LOG(4, ("domain_init: regxep '%s'", result->pattern));
  
  regex_t preg;
  if (regcomp(&result->preg, result->pattern, REG_EXTENDED|REG_ICASE) != 0) {
    warnx("Failed to compile '%s'", result->pattern);
    goto compile_failed;
  }
  result->nmatch = groups + 1;

  result->user_path = strdup_or_null(config_get("Files", "User-", domain));
  if (!result->user_path) {
    warnx("files_init: no 'User-%s' specified", domain);
    goto no_user_file;
  }
  if (access(result->user_path, R_OK) != 0) {
    warn("files_init: '%s' not readable", result->user_path);
    goto no_user_file_access;
  }
  result->group_path = strdup_or_null(config_get("Files", "Group-", domain));
  if (!result->group_path) {
    warnx("files_init: no 'Group-%s' specified", domain);
    goto no_group_file;
  }
  if (access(result->group_path, R_OK) != 0) {
    warn("files_init: '%s' not readable", result->group_path);
    goto no_group_file_access;
  }
  result->name = strdup_or_null(config_get("Files", "Domain-", domain));
  if (!result->name) {
    warnx("files_init: no 'Domain-%s' specified", domain);
    goto no_domain_name;
  }
  free(patterns);  
  return result;

no_domain_name:
no_group_file_access:
  free(result->group_path);
no_group_file:
no_user_file_access:
  free(result->user_path);
no_user_file:
  regfree(&result->preg);
compile_failed:
invalid_pattern:  
  free(patterns);  
no_patterns:
  free(result);
no_result:
  return NULL;
}
  
static int files_init() {	
  int status;

  char *domains, *domain, *domain_save;
  struct domain **tail = &files_domain;
  
  domains = strdup_or_null(nfsidmap_config_get("Files", "Domains"));
  if (!domains) {
    warnx("files_init: Domains for user mapping missing");
    goto no_domains;
  }
  IDMAP_LOG(4, ("files_init: found domains %s", domains));
  for (domain = strtok_r(domains, ",", &domain_save) ;
       domain ;
       domain = strtok_r(NULL, ",", &domain_save)) {
    struct domain *d = domain_init(domain);
    if (!d) {
      goto domain_new_failed;
    }
    *tail = d;
    tail = &d->next;
  }
  free(domains);
  
  return 0;
  
domain_new_failed:
no_domains:
  return -EINVAL;
}

static int files_gss_princ_to_ids(char *secname, char *princ,
                                  uid_t *uid, uid_t *gid,
                                  extra_mapping_params **ex)
{
  int err;
  struct passwd pw;
  void *when_ready = NULL;

  err = files_getpwnam(&pw, princ, NULL, NULL, &when_ready);
  if (err == 0) {
    *uid = pw.pw_uid;
    *gid = pw.pw_gid;
  }
  free(when_ready);
  return -err;
}

static int files_gss_princ_to_grouplist(char *secname, char *princ,
                                        gid_t *groups, int *ngroups,
                                        extra_mapping_params **ex)
{
  int err;
  struct passwd pw;
  struct domain *domain;
  void *when_ready = NULL;
  
  err = files_getpwnam(&pw, princ, NULL, &domain, &when_ready);
  if (err == 0) {
    if (files_getgrouplist(domain, pw.pw_name, pw.pw_gid,
                           groups, ngroups) < 0) {
      err = ERANGE;
    }
  }
  free(when_ready);
  return err;
}

struct trans_func files_trans = {
	.name			= "files",
	.init			= files_init,
	.name_to_uid		= files_name_to_uid,
	.name_to_gid		= files_name_to_gid,
	.uid_to_name		= files_uid_to_name,
	.gid_to_name		= files_gid_to_name,
	.princ_to_ids		= files_gss_princ_to_ids,
	.gss_princ_to_grouplist	= files_gss_princ_to_grouplist,
};

struct trans_func *libnfsidmap_plugin_init()
{
	return (&files_trans);
}

